#
#  CNWindow.rb
#  CocoaNav
#
#  Created by Patrick Geiller on 09/05/08.
#  Copyright (c) 2008 __MyCompanyName__. All rights reserved.
#

require_framework 'ScriptingBridge'

#
# CNWindow is the main window class
#	coordinates between FrameworkView, CrumbView, ItemTableView to share their selection
#
class CNWindow < NSWindow

	
	ib_outlet :document
	
	# Toolbar items
	ib_outlet :toolbarNewItem
	ib_outlet :toolbarViewMode
	ib_outlet :toolbarZoom
	ib_outlet :toolbarHelp
	ib_outlet :toolbarSearch
	kvc_accessor :toolbarNewItem
	kvc_accessor :toolbarViewMode
	kvc_accessor :toolbarZoom
	kvc_accessor :toolbarHelp
	kvc_accessor :toolbarSearch

	ib_outlet :frameworkView
	ib_outlet :crumbView
	ib_outlet :tableView
	ib_outlet :webView
	ib_outlet :splitView
	ib_outlet :helpSplitView


	# Array controller
	ib_outlet :itemList
	kvc_accessor :itemList
	# Shared used defaults
	ib_outlet :userDefaults

	#
	# awakeFromNib
	#
	def awakeFromNib

		# Default filter predicate for items
		@predicate = 'origin == "sys"'

		# Link to document
		@documentType = nil
		self.bind_toObject_withKeyPath_options('documentType', @document, 'type', nil);

		# Selected class
		@value = ''
		@loadedFrameworks = false

		@appController = NSApp.delegate
		@appController.addObserver_forKeyPath_options_context(self, 'items', 0, nil)


		setupToolbar

		@numFrameworksLoaded = 0
		
		@frameworkView.setSource(NSApp.delegate)
		@crumbView.setSource(NSApp.delegate)
		@tableView.setSource(self)

		# Bind background colors to frameworkView
		@frameworkView.bind_toObject_withKeyPath_options(	'backgroundColor1', 
															@userDefaults, 'values.backgroundColor1', 
															NSDictionary.dictionaryWithObjectsAndKeys(NSUnarchiveFromDataTransformerName, 'NSValueTransformerName', nil ));

		@frameworkView.bind_toObject_withKeyPath_options(	'backgroundColor2', 
															@userDefaults, 'values.backgroundColor2', 
															NSDictionary.dictionaryWithObjectsAndKeys(NSUnarchiveFromDataTransformerName, 'NSValueTransformerName', nil ));

		# Coordinates between all views to synch their selected class
		@frameworkView.addObserver_forKeyPath_options_context(self, 'value', 0, nil)
		@crumbView.addObserver_forKeyPath_options_context(self, 'value', 0, nil)
		@tableView.addObserver_forKeyPath_options_context(self, 'value', 0, nil)
		self.addObserver_forKeyPath_options_context(self, 'value', 0, nil)
		NSApp.delegate.addObserver_forKeyPath_options_context(self, 'colorChanged', 0, nil)


		# How much hep pane is open relative to frame.width
		@helpPaneSizeRatio = 0.7

		setDelegate self
	end

	#
	# dealloc
	#
	def dealloc
		@frameworkView.removeObserver_forKeyPath(self, 'value')
		@crumbView.removeObserver_forKeyPath(self, 'value')
		@tableView.removeObserver_forKeyPath(self, 'value')
		self.removeObserver_forKeyPath(self, 'value')
		
		super_dealloc
	end
	
	

	
	#
	# KVO
	#	Called from Application Controller when adding an item, dispatches the call to FrameworkView
	#	Called from FrameworkView, CrumbView, ItemTableView when selection changes
	#	Called from ApplicationController when color changes
	#
	def observeValueForKeyPath_ofObject_change_context(keyPath, object, change, context)

		# Shared value (selected class) among views
		if keyPath != 'items'
			if keyPath == 'value'
				value = object.valueForKey('value')
				@value = value

				@crumbView.setValue_forKey(value, 'value') if @crumbView.valueForKey('value') != value
				@tableView.setValue_forKey(value, 'value') if @tableView.valueForKey('value') != value
				@frameworkView.setValue_forKey(value, 'value') if @frameworkView.valueForKey('value') != value
				reflectSelectedClassInHelp
				
			end
			if keyPath == 'colorChanged'
				@frameworkView.updateColors if @frameworkView
			end
			
			return
		end

		# Add frameworks and classes to framework view
		idx = @loadItemNumber || @appController.items.length-1
		obj = @appController.items[idx]
		return	if obj == nil

		# Get out if it's a user framework that doesn't belong to us
		return	if obj.origin == 'user' && !@document.isUserFrameworkLoaded(obj.framework)
		# Get out if it's a sys framework and we've already loaded frameworks
		return	if obj.origin == 'sys' and @loadedFrameworks
		# Get out if it's a sys framework and we're not a sys window
		return	if obj.origin == 'sys' and @documentType == 'blank'

		loadedFrameworks if obj.oc_type == 'end'

		if obj.oc_type == 'framework'
		@itemList.setFilterPredicate(nil)
			@frameworkView.addFramework(obj.framework, obj.origin)
			@numFrameworksLoaded += 1
		elsif obj.oc_type == 'class'
			@frameworkView.addClass(obj.name)		
		end
	end
	
	#
	# Loaded frameworks
	#	notify framework view
	#	enable toolbar items
	#	
	def	loadedFrameworks
		@frameworkView.loadedFrameworks
		enableToolbarItems(true)
		@loadedFrameworks = true
		setFilterPredicate('')
	end


	#
	# Selected class
	#
	kvc_accessor :value
	kvc_accessor :documentType
	
	#
	# setDocumentType
	#
	def	setValue_forKey(v, k)
		super_setValue_forKey(v, k)
		if k == 'documentType'
			if v == 'blank'
				@predicate = 'origin == "none"'
				setFilterPredicate('')
			end	
			reloadItems
			loadedFrameworks
		end
	end

	#
	# Go through items to add them in KVO observe method
	#
	def	reloadItems
		# Load frameworks
		@loadItemNumber = 0
		NSApp.delegate.items.each { |item|
			next if @document.isUserFrameworkLoaded(item.framework)
			observeValueForKeyPath_ofObject_change_context('items', nil, nil, nil)
			@loadItemNumber += 1
		}
		@loadItemNumber = nil
	end




	#
	# Update table view's filter predicate
	#
	def	setFilterPredicate(term)
		# Split search query to get search terms, build predicate like 
		# (searchName contains %@) and (searchName contains %@) and (searchName contains %@), [arguments array]
		terms = term.split(' ')
		p = ''
		if (terms.length > 0)
			ps = [];	terms.length.times { ps << '(searchName contains[c] %@)' }
			p = ps.join(' and ')
			p += ' and '
		end

		# Handle user frameworks predicate
		if @document.userFrameworks.length > 0
			p += '  (' + @predicate
			@document.userFrameworks.each { |fw|
				p += ' or framework = %@'
				terms << fw
			}
			p += ')'
		elsif
			p += '  (' + @predicate + ')'
		end

		predicate = NSPredicate.predicateWithFormat_argumentArray(p, terms)
#		puts 'setFilterPredicate=' + predicate.to_s
		@itemList.setFilterPredicate(predicate)

		# If no selection, select first item
		@itemList.setSelectionIndex(0) if @itemList.selectionIndex == NSNotFound
	end
	
	
	def	itemFromSelectedClass
		NSApp.delegate.items.each { |item|
			return item if item.oc_type == 'class' && item.name == @value
		}
		nil
	end



	#
	#
	#	Toolbar
	#
	#
	
	#
	# Setup toolbar
	#
	def setupToolbar
		# Setup toolbar
		@toolbarItems = [	'toolbarNewItem', 
							NSToolbarFlexibleSpaceItemIdentifier, 
							'toolbarViewMode', 
							NSToolbarSpaceItemIdentifier, 
							'toolbarZoom', 
							NSToolbarSpaceItemIdentifier, 
							'toolbarHelp', 
							NSToolbarFlexibleSpaceItemIdentifier, 
							'toolbarSearch']
							
		
		@toolbarViewMode.imageForSegment(0).setTemplate(true)
		@toolbarViewMode.imageForSegment(1).setTemplate(true)

		@toolbar = NSToolbar.alloc.initWithIdentifier 'CNToolbar'
		@toolbar.setDelegate self
		@toolbar.setAllowsUserCustomization false
		self.setToolbar @toolbar
		enableToolbarItems(false)
	end
	
	#
	# Enable / disable all toolbar items
	#
	def	enableToolbarItems(b)
		items = @toolbar.items
		items.each { |item|
			item.setEnabled(b)
		}
	end

	#
	# Toolbar allowed identifiers
	#
	def toolbarAllowedItemIdentifiers(toolbar)
		return	@toolbarItems
	end

	#
	# Toolbar default items
	#
	def toolbarDefaultItemIdentifiers(toolbar)
		return	@toolbarItems
	end

	#
	# Toolbar instance for identifier
	#
	def toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar(toolbar, itemIdentifier, flag)
		toolbarItem = NSToolbarItem.alloc.initWithItemIdentifier(itemIdentifier)
		
		item = valueForKey(itemIdentifier)
		label = item.nextKeyView ? item.nextKeyView.stringValue : ''
		toolbarItem.setLabel label
		toolbarItem.setView(item)

		return	toolbarItem
	end



	
	#
	#
	# Sheet : open framework
	#	accept folders and .xcodeproj files
	#
	#
	ib_action :openFramework
	def openFramework(sender)
		panel = NSOpenPanel.openPanel
		panel.setCanChooseFiles(true)
		panel.setCanChooseDirectories(true)
		panel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, ['xcodeproj'], OSX::NSApp.mainWindow, self, 'openPanelDidEnd_returnCode_contextInfo', nil)
	end
	
	#
	# Return method from sheet
	#
	def openPanelDidEnd_returnCode_contextInfo(panel, returnCode, contextInfo)
		if (returnCode == NSOKButton)
			@document.maybeLoadUserFramework(panel.filenames[0])
			@frameworkView.loadedFrameworks
		end
	end
	





	#
	#
	# IBActions
	#
	#


	#
	# Find
	#
	def	find(sender)
		self.makeFirstResponder(@toolbarSearch)
	end
	
	#
	# Find next
	#	(go to next row of table view)
	#
	def	findNext(sender)
		row = @tableView.selectedRow+1
		# Wrap to first
		row = 0 if (row >= @tableView.numberOfRows)
		
		@tableView.selectRowIndexes_byExtendingSelection(NSIndexSet.indexSetWithIndex(row), false)
		@tableView.scrollRowToVisible(row)
	end
	
	#
	# Find previous
	#
	def	findPrevious(sender)
		row = @tableView.selectedRow-1
		# Wrap to last
		row = @tableView.numberOfRows-1 if (row < 0)

		@tableView.selectRowIndexes_byExtendingSelection(NSIndexSet.indexSetWithIndex(row), false)
		@tableView.scrollRowToVisible(row)
	end

	ib_action :toolbarSearchChanged
	def toolbarSearchChanged(sender)
		setFilterPredicate(sender.stringValue)	
	end

	def	openInFinder(sender)
		item = itemFromSelectedClass
		return if item == nil

		system('open "' + item.file + '"')
	end
	
	def	revealInFinder(sender)
		item = itemFromSelectedClass
		return if item == nil

		NSWorkspace.sharedWorkspace.selectFile_inFileViewerRootedAtPath(item.file, nil)
	end

	def openInCocoaDev(sender)
		item = itemFromSelectedClass
		return if item == nil

		c = item.name if item.oc_type == 'class'
		c = item.associatedClass if item.associatedClass
		
		return if c == nil

		url = 'http://www.cocoadev.com/index.pl?' + c
		system('open ' + url)
	end


	def helpFileOfClass(c)
		notFound = NSBundle.mainBundle.pathForResource_ofType('helpNotFound', 'html')
		return	notFound if c == nil || c.to_s.length == 0

		#
		# Help for NSSomeClass is usually stored in a folder as /NSSomeClass_class/index.html
		# But some are stored as /NSSomeClass/index.html
		#
		# Use spotlight to find our search term in Developer Doc
		# Then match result with grep, finding the line that ends ($) with our search term (and possibly with _class, or _protocol)
		#
		path = `mdfind -onlyin /Developer/Documentation #{c} | grep -iE '#{c}(_[A-Za-z]+)?$'`
		return	notFound if path.length == 0

		# Keep only first result
		path = path.split("\n")[0]
		path + '/index.html'
	end
		
	def openHelpInSafari(sender)
		item = itemFromSelectedClass
		return if item == nil

		c = item.name if item.oc_type == 'class'
		c = item.associatedClass if item.associatedClass

		file = helpFileOfClass(c)
		system('open "' + file + '"')
	end


	# Copy class name
	def copy(sender)
		return if @tableView.selectedRow == -1
		str = @itemList.arrangedObjects.objectAtIndex(@tableView.selectedRow).name
		NSPasteboard.generalPasteboard.setString_forType(str, NSStringPboardType)
	end
	


	#
	#
	# Class Tree walking
	#
	#

	def goToNSObject(sender)
		setValue_forKey('NSObject', 'value') 
	end
	def goToParent(sender)
		c = @appController.classParent(@value)
		return if c == nil
		setValue_forKey(c, 'value') 
	end
	def goToNextSibling(sender)
		c = @appController.classNextSiblingWrap(@value)
		return if c == nil
		setValue_forKey(c, 'value') 
	end
	def goToPreviousSibling(sender)
		c = @appController.classPreviousSiblingWrap(@value)
		return if c == nil
		setValue_forKey(c, 'value') 
	end
	def goToFirstChild(sender)
		c = @appController.classFirstChild(@value)
		return if c == nil
		setValue_forKey(c, 'value') 
	end
	
	

	def goToNextInTree(sender)
		c = @appController.classNext(@value)
		setValue_forKey(c, 'value') 
	end
	def goToPreviousInTree(sender)
		c = @appController.classPrevious(@value)
		setValue_forKey(c, 'value') 
	end

	def	switchToFrameworkView(sender)
		@frameworkView.switchToFrameworkView
	end
	def	switchToClassTreeView(sender)
		@frameworkView.switchToClassHierarchyView
	end
	
	ib_action :zoomIn
	def zoomIn
		@frameworkView.zoomIn
	end
	ib_action :zoomOut
	def zoomOut
		@frameworkView.zoomOut
	end
	ib_action :zoomExtent
	def zoomExtent
		@frameworkView.zoomExtent
	end
	


	#
	#
	# Split View
	#
	#

	#
	# Constrain horizontal split view to size of crumb view
	#
	def splitView_constrainSplitPosition_ofSubviewAt(splitView, position, index)
		# If it's the help splitView, record position and toggle help button
		if splitView.to_s == @helpSplitView.to_s
			@helpPaneSizeRatio = 1 - (position / frame.width)
			
			state = @helpPaneSizeRatio > 0 ? 1 : 0
			if state != @toolbarHelp.state
				@toolbarHelp.setState(state)
				
				reflectSelectedClassInHelp
			end
			
			return position
		end
		heightOfDividerAndBottomView = splitView.dividerThickness + 24 - 3
		[position, splitView.frame.height-heightOfDividerAndBottomView].min
	end
	
	#
	# Make sure crumbview does not get squished during resize
	#
	def	splitViewDidResizeSubviews(notif)
	
		view = @splitView.subviews[1]
		return if view == nil
		min = 24
		h = view.frame.height - min
		if view.frame.height <= min
			view.setFrameSize(NSMakeSize(frame.width, min))
			h = 0
		end
		
		@tableView.superview.superview.setHidden(h < 2)
		
		@crumbView.setFrameSize(NSMakeSize(frame.width, min))
		v = @tableView.superview.superview
		v.setFrame(NSMakeRect(0, 0, v.frame.width, h))
		@splitView.adjustSubviews
	end
	



	#
	# Disallow any menu item until we've loaded frameworks
	#
	def validateMenuItem(item)
		return	@loadedFrameworks
	end

	#
	# When selecting a class, find its help and show it in WebView
	#
	def reflectSelectedClassInHelp
		return if @toolbarHelp.state == 0

		return if @lastClassOpenInWebView == @value

		file = helpFileOfClass(@value)
		@lastClassOpenInWebView = @value
		
		@webView.mainFrame.loadRequest(NSURLRequest.requestWithURL(NSURL.fileURLWithPath(file)))
	end
	
	
	ib_action :toggleHelp
	def toggleHelp(sender)
		paneSize = (1-@helpPaneSizeRatio)*frame.width
		# Minimal size
		paneSize = [paneSize, frame.width - 100].min

		state = @toolbarHelp.state
		# Special case for menu item activation
		if sender.class.to_s.match(/menu/i)
			state = state == 0 ? 1 : 0
		end

		size = state == 1 ? paneSize : frame.width

		# Save ratio which will be overwritten in the delegate method
		s = @helpPaneSizeRatio
		# Set position
		@helpSplitView.setPosition_ofDividerAtIndex(size, 0)
		@helpPaneSizeRatio = s
		
		reflectSelectedClassInHelp
	end
	


	#
	# Full screen handler (will handle fullscreen if framework view hasn't the focus
	#
	def	toggleFullScreen(sender)
		@frameworkView.toggleFullScreen(sender)
	end
	
end
