#
#  CrumbView.rb
#  CocoaNav
#
#  Created by Patrick Geiller on 11/05/08.
#  Copyright (c) 2008 __MyCompanyName__. All rights reserved.
#

require 'osx/cocoa'

#
# CrumView draws all ancestors of a selected class
#
class CrumbView <  NSView

	def awakeFromNib
		@shadow = NSShadow.alloc.init
		@shadow.shadowOffset = NSMakeSize(0, -1)
		@shadow.shadowBlurRadius = 1
		@shadow.shadowColor = NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0.75)
		
		@positions = []
		
		@value = ''
	end

	def setSource(source)
		@source = source
	end

	def drawRect(rect)

		# Draw background
		brightness = 0.3
		NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 1).set
		NSBezierPath.bezierPathWithRect(rect).fill

		return	if @source == nil
		
		hash = @source.hashChildren
#		@selectedClass = 'NSObject' if @selectedClass == nil
		return if @selectedClass == nil
		
		# Build derivation crumb
		c = @selectedClass
		items = []
		while c != nil && c != ''
			items << c
			c = @source.parentClassOfClass(c.to_s)
		end
		items.reverse!

		# Setup text colors
		fontSize = 12
		attributes1 = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.whiteColor, NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																@shadow, NSShadowAttributeName,
													nil);
		attributes2 = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithCalibratedWhite_alpha(0.6, 1), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																@shadow, NSShadowAttributeName,
													nil);

		# Center in view
		delta = (rect.size.height - NSString.stringWithString('A').sizeWithAttributes(attributes1).height)/2 - 0.5
		rect.y -= delta
		
		# Print each class name followed by an arrow
		items.each_with_index { |item, index|
		
			attributes = index == items.length-1 ? attributes1 : attributes2
		
			str = NSString.stringWithString(' ' + item + ' ')
			str.drawInRect_withAttributes(rect, attributes)
			width = str.sizeWithAttributes(attributes).width
			@positions << { 'x1' => rect.x, 'x2' => rect.x + width, 'className' => item, 'type' => 'class' }
			rect.x += width

			str = NSString.stringWithString(' → ')
			str.drawInRect_withAttributes(rect, attributes2)
			width = str.sizeWithAttributes(attributes).width
			@positions << { 'x1' => rect.x, 'x2' => rect.x + width, 'className' => item, 'type' => 'children' }
			rect.x += width
		}
		
	end

	def mouseDown(event)
		point = convertPoint_fromView(event.locationInWindow, nil)

		# Determine which position we clicked on
		clickedPosition = nil
		@positions.each { |position|
			if point.x >= position['x1'] && point.x <= position['x2']
				clickedPosition = position
			end
		}
		return if clickedPosition == nil

		# Click on a class, get siblings
		# Click on arrow, get children
		if (clickedPosition['type'] == 'class')
			parentClass = @source.parentClassOfClass(clickedPosition['className'])
		else
			parentClass = clickedPosition['className']
		end
		
		# Get object list to display
		if parentClass == nil
			items = @source.rootObjects
		else
			items = @source.hashChildren[parentClass.to_s]
		end

		# Reset menu by removing any previously inserted class
		while menu.numberOfItems > 1 && menu.itemAtIndex(1)
			menu.removeItemAtIndex(1)
		end
		
		# Populate menu
		items = [] if items == nil
		items.each { |item|
			item = NSMenuItem.alloc.initWithTitle_action_keyEquivalent(item, :selectSiblingInCrumbView, '')
			menu.addItem item		
		}
		
		# Show or hide 'No children' menu item
		menu.itemAtIndex(0).setHidden(items.length > 0)

		# Show menu
		NSMenu.popUpContextMenu_withEvent_forView(self.menu, event, self)
	end
	
	def rightMouseDown(event)
		mouseDown(event)
	end

	
	def selectSiblingInCrumbView(sender)
		select(sender.title)
	end
	
	def select(name)
		@selectedClass = name
		setNeedsDisplay(true)
		setValue_forKey(@selectedClass, 'value')
	end


	#
	# Selected class
	#
	kvc_accessor :value
	def setValue_forKey(value, key)
#		puts 'from CrumbView:' + value.to_s + '=' + key + '!'
		
		if key == 'value'
			return if value == @value
			self.performSelector_withObject_afterDelay(:updateWithNewValue, nil, 0)
		end
		
		super_setValue_forKey(value, key)
	end
	
	def updateWithNewValue
		select(@value)
	end
end
